/**
 *
 * @file spm_check_and_correct_tests.c
 *
 * Tests and validate the spmCheckAndCorrect routines.
 *
 * @copyright 2015-2022 Bordeaux INP, CNRS (LaBRI UMR 5800), Inria,
 *                      Univ. Bordeaux. All rights reserved.
 *
 * @version 1.1.1
 * @author Mathieu Faverge
 * @author Tony Delarue
 * @date 2021-12-06
 *
 **/
#include <stdint.h>
#include <math.h>
#include <time.h>
#include <spm_tests.h>

static inline int
spm_check_and_correct_check_merge_duplicate(const spmatrix_t *spm)
{
    spm_int_t  i, j;
    spm_int_t  size, n;
    spm_int_t *colptr = (spm->fmttype == SpmCSC) ? spm->colptr : spm->rowptr;
    spm_int_t *rowptr = (spm->fmttype == SpmCSC) ? spm->rowptr : spm->colptr;

    n       = spm->n;
    for ( j = 0; j < n; j++, colptr++)
    {
        size = colptr[1] - colptr[0] - 1;
        for ( i = 0; i < size; i++, rowptr++ )
        {
            /* MergeDuplicate should have been called */
            if( rowptr[0] == rowptr[1] ) {
                return 1;
            }
        }
        rowptr++;
    }
    return 0;
}

static inline int
spm_check_and_correct_check_symmetrize(const spmatrix_t *spm)
{
    spm_int_t  i, j, col, row, n, baseval;
    spm_int_t  found, index, size;
    spm_int_t *colptr = (spm->fmttype == SpmCSC) ? spm->colptr : spm->rowptr;
    spm_int_t *coltmp = colptr;
    spm_int_t *rowptr = (spm->fmttype == SpmCSC) ? spm->rowptr : spm->colptr;
    spm_int_t *rowtmp = rowptr;

    if ( spm->mtxtype != SpmGeneral ) {
        return 0;
    }

    n       = spm->n;
    baseval = spmFindBase(spm);
    for ( col = 0; col < n; col++, coltmp++)
    {
        for ( i = coltmp[0]; i < coltmp[1]; i++, rowtmp++ )
        {
            row = *rowtmp - baseval;

            index = colptr[row] - baseval;
            size  = colptr[row + 1] - colptr[row];
            /* Check the symmetry */
            for ( j = 0; j < size; j++)
            {
                found = rowptr[index + j] - baseval;
                if( found == col ) {
                    break;
                }
                /* We've sort the matrix */
                if( found > col ) {
                    return 1;
                }
            }
        }
    }
    return 0;
}

static inline int
spm_check_and_correct_check( const spmatrix_t *spm )
{
    spmatrix_t spm_out;
    int        rc1 = 0, rc2 = 0;
    int        new;

    new = spmCheckAndCorrect( spm, &spm_out );
    if( new ) {
        assert(spm_out.fmttype == SpmCSC);
        /* Sort it */
        spmSort( &spm_out );

        rc1 = spm_check_and_correct_check_merge_duplicate( &spm_out );
        rc2 = spm_check_and_correct_check_symmetrize( &spm_out );

        spmExit( &spm_out );
    }

    return rc1 + rc2;
}

int main (int argc, char **argv)
{
    spmatrix_t original;
    int        rc, err = 0;

#if defined(SPM_WITH_MPI)
    MPI_Init( &argc, &argv );
#endif

    /**
     * Get options from command line
     */
    rc = spmTestGetSpm( &original, argc, argv );

    if ( rc != SPM_SUCCESS ) {
        fprintf(stderr, "ERROR: Could not read the file, stop the test !!!\n");
        return EXIT_FAILURE;
    }

    printf(" -- SPM CheckAndCorrect Test --\n");
    err = spmTestLoop( &original, &spm_check_and_correct_check, 0 );

    spmExit( &original  );

#if defined(SPM_WITH_MPI)
    MPI_Finalize();
#endif

    return spmTestEnd( err, 0 );
}
